// generated by go run gen.go; DO NOT EDIT

package atom

const (
	A                   Atom = 0x1
	Abbr                Atom = 0x4
	Accept              Atom = 0x2106
	AcceptCharset       Atom = 0x210e
	Accesskey           Atom = 0x3309
	Action              Atom = 0x1f606
	Address             Atom = 0x4f307
	Align               Atom = 0x1105
	Alt                 Atom = 0x4503
	Annotation          Atom = 0x1670a
	AnnotationXml       Atom = 0x1670e
	Applet              Atom = 0x2b306
	Area                Atom = 0x2fa04
	Article             Atom = 0x38807
	Aside               Atom = 0x8305
	Async               Atom = 0x7b05
	Audio               Atom = 0xa605
	Autocomplete        Atom = 0x1fc0c
	Autofocus           Atom = 0xb309
	Autoplay            Atom = 0xce08
	B                   Atom = 0x101
	Base                Atom = 0xd604
	Basefont            Atom = 0xd608
	Bdi                 Atom = 0x1a03
	Bdo                 Atom = 0xe703
	Bgsound             Atom = 0x11807
	Big                 Atom = 0x12403
	Blink               Atom = 0x12705
	Blockquote          Atom = 0x12c0a
	Body                Atom = 0x2f04
	Br                  Atom = 0x202
	Button              Atom = 0x13606
	Canvas              Atom = 0x7f06
	Caption             Atom = 0x1bb07
	Center              Atom = 0x5b506
	Challenge           Atom = 0x21f09
	Charset             Atom = 0x2807
	Checked             Atom = 0x32807
	Cite                Atom = 0x3c804
	Class               Atom = 0x4de05
	Code                Atom = 0x14904
	Col                 Atom = 0x15003
	Colgroup            Atom = 0x15008
	Color               Atom = 0x15d05
	Cols                Atom = 0x16204
	Colspan             Atom = 0x16207
	Command             Atom = 0x17507
	Content             Atom = 0x42307
	Contenteditable     Atom = 0x4230f
	Contextmenu         Atom = 0x3310b
	Controls            Atom = 0x18808
	Coords              Atom = 0x19406
	Crossorigin         Atom = 0x19f0b
	Data                Atom = 0x44a04
	Datalist            Atom = 0x44a08
	Datetime            Atom = 0x23c08
	Dd                  Atom = 0x26702
	Default             Atom = 0x8607
	Defer               Atom = 0x14b05
	Del                 Atom = 0x3ef03
	Desc                Atom = 0x4db04
	Details             Atom = 0x4807
	Dfn                 Atom = 0x6103
	Dialog              Atom = 0x1b06
	Dir                 Atom = 0x6903
	Dirname             Atom = 0x6907
	Disabled            Atom = 0x10c08
	Div                 Atom = 0x11303
	Dl                  Atom = 0x11e02
	Download            Atom = 0x40008
	Draggable           Atom = 0x17b09
	Dropzone            Atom = 0x39108
	Dt                  Atom = 0x50902
	Em                  Atom = 0x6502
	Embed               Atom = 0x6505
	Enctype             Atom = 0x21107
	Face                Atom = 0x5b304
	Fieldset            Atom = 0x1b008
	Figcaption          Atom = 0x1b80a
	Figure              Atom = 0x1cc06
	Font                Atom = 0xda04
	Footer              Atom = 0x8d06
	For                 Atom = 0x1d803
	ForeignObject       Atom = 0x1d80d
	Foreignobject       Atom = 0x1e50d
	Form                Atom = 0x1f204
	Formaction          Atom = 0x1f20a
	Formenctype         Atom = 0x20d0b
	Formmethod          Atom = 0x2280a
	Formnovalidate      Atom = 0x2320e
	Formtarget          Atom = 0x2470a
	Frame               Atom = 0x9a05
	Frameset            Atom = 0x9a08
	H1                  Atom = 0x26e02
	H2                  Atom = 0x29402
	H3                  Atom = 0x2a702
	H4                  Atom = 0x2e902
	H5                  Atom = 0x2f302
	H6                  Atom = 0x50b02
	Head                Atom = 0x2d504
	Header              Atom = 0x2d506
	Headers             Atom = 0x2d507
	Height              Atom = 0x25106
	Hgroup              Atom = 0x25906
	Hidden              Atom = 0x26506
	High                Atom = 0x26b04
	Hr                  Atom = 0x27002
	Href                Atom = 0x27004
	Hreflang            Atom = 0x27008
	Html                Atom = 0x25504
	HttpEquiv           Atom = 0x2780a
	I                   Atom = 0x601
	Icon                Atom = 0x42204
	Id                  Atom = 0x8502
	Iframe              Atom = 0x29606
	Image               Atom = 0x29c05
	Img                 Atom = 0x2a103
	Input               Atom = 0x3e805
	Inputmode           Atom = 0x3e809
	Ins                 Atom = 0x1a803
	Isindex             Atom = 0x2a907
	Ismap               Atom = 0x2b005
	Itemid              Atom = 0x33c06
	Itemprop            Atom = 0x3c908
	Itemref             Atom = 0x5ad07
	Itemscope           Atom = 0x2b909
	Itemtype            Atom = 0x2c308
	Kbd                 Atom = 0x1903
	Keygen              Atom = 0x3906
	Keytype             Atom = 0x53707
	Kind                Atom = 0x10904
	Label               Atom = 0xf005
	Lang                Atom = 0x27404
	Legend              Atom = 0x18206
	Li                  Atom = 0x1202
	Link                Atom = 0x12804
	List                Atom = 0x44e04
	Listing             Atom = 0x44e07
	Loop                Atom = 0xf404
	Low                 Atom = 0x11f03
	Malignmark          Atom = 0x100a
	Manifest            Atom = 0x5f108
	Map                 Atom = 0x2b203
	Mark                Atom = 0x1604
	Marquee             Atom = 0x2cb07
	Math                Atom = 0x2d204
	Max                 Atom = 0x2e103
	Maxlength           Atom = 0x2e109
	Media               Atom = 0x6e05
	Mediagroup          Atom = 0x6e0a
	Menu                Atom = 0x33804
	Menuitem            Atom = 0x33808
	Meta                Atom = 0x45d04
	Meter               Atom = 0x24205
	Method              Atom = 0x22c06
	Mglyph              Atom = 0x2a206
	Mi                  Atom = 0x2eb02
	Min                 Atom = 0x2eb03
	Minlength           Atom = 0x2eb09
	Mn                  Atom = 0x23502
	Mo                  Atom = 0x3ed02
	Ms                  Atom = 0x2bc02
	Mtext               Atom = 0x2f505
	Multiple            Atom = 0x30308
	Muted               Atom = 0x30b05
	Name                Atom = 0x6c04
	Nav                 Atom = 0x3e03
	Nobr                Atom = 0x5704
	Noembed             Atom = 0x6307
	Noframes            Atom = 0x9808
	Noscript            Atom = 0x3d208
	Novalidate          Atom = 0x2360a
	Object              Atom = 0x1ec06
	Ol                  Atom = 0xc902
	Onabort             Atom = 0x13a07
	Onafterprint        Atom = 0x1c00c
	Onautocomplete      Atom = 0x1fa0e
	Onautocompleteerror Atom = 0x1fa13
	Onbeforeprint       Atom = 0x6040d
	Onbeforeunload      Atom = 0x4e70e
	Onblur              Atom = 0xaa06
	Oncancel            Atom = 0xe908
	Oncanplay           Atom = 0x28509
	Oncanplaythrough    Atom = 0x28510
	Onchange            Atom = 0x3a708
	Onclick             Atom = 0x31007
	Onclose             Atom = 0x31707
	Oncontextmenu       Atom = 0x32f0d
	Oncuechange         Atom = 0x3420b
	Ondblclick          Atom = 0x34d0a
	Ondrag              Atom = 0x35706
	Ondragend           Atom = 0x35709
	Ondragenter         Atom = 0x3600b
	Ondragleave         Atom = 0x36b0b
	Ondragover          Atom = 0x3760a
	Ondragstart         Atom = 0x3800b
	Ondrop              Atom = 0x38f06
	Ondurationchange    Atom = 0x39f10
	Onemptied           Atom = 0x39609
	Onended             Atom = 0x3af07
	Onerror             Atom = 0x3b607
	Onfocus             Atom = 0x3bd07
	Onhashchange        Atom = 0x3da0c
	Oninput             Atom = 0x3e607
	Oninvalid           Atom = 0x3f209
	Onkeydown           Atom = 0x3fb09
	Onkeypress          Atom = 0x4080a
	Onkeyup             Atom = 0x41807
	Onlanguagechange    Atom = 0x43210
	Onload              Atom = 0x44206
	Onloadeddata        Atom = 0x4420c
	Onloadedmetadata    Atom = 0x45510
	Onloadstart         Atom = 0x46b0b
	Onmessage           Atom = 0x47609
	Onmousedown         Atom = 0x47f0b
	Onmousemove         Atom = 0x48a0b
	Onmouseout          Atom = 0x4950a
	Onmouseover         Atom = 0x4a20b
	Onmouseup           Atom = 0x4ad09
	Onmousewheel        Atom = 0x4b60c
	Onoffline           Atom = 0x4c209
	Ononline            Atom = 0x4cb08
	Onpagehide          Atom = 0x4d30a
	Onpageshow          Atom = 0x4fe0a
	Onpause             Atom = 0x50d07
	Onplay              Atom = 0x51706
	Onplaying           Atom = 0x51709
	Onpopstate          Atom = 0x5200a
	Onprogress          Atom = 0x52a0a
	Onratechange        Atom = 0x53e0c
	Onreset             Atom = 0x54a07
	Onresize            Atom = 0x55108
	Onscroll            Atom = 0x55f08
	Onseeked            Atom = 0x56708
	Onseeking           Atom = 0x56f09
	Onselect            Atom = 0x57808
	Onshow              Atom = 0x58206
	Onsort              Atom = 0x58b06
	Onstalled           Atom = 0x59509
	Onstorage           Atom = 0x59e09
	Onsubmit            Atom = 0x5a708
	Onsuspend           Atom = 0x5bb09
	Ontimeupdate        Atom = 0xdb0c
	Ontoggle            Atom = 0x5c408
	Onunload            Atom = 0x5cc08
	Onvolumechange      Atom = 0x5d40e
	Onwaiting           Atom = 0x5e209
	Open                Atom = 0x3cf04
	Optgroup            Atom = 0xf608
	Optimum             Atom = 0x5eb07
	Option              Atom = 0x60006
	Output              Atom = 0x49c06
	P                   Atom = 0xc01
	Param               Atom = 0xc05
	Pattern             Atom = 0x5107
	Ping                Atom = 0x7704
	Placeholder         Atom = 0xc30b
	Plaintext           Atom = 0xfd09
	Poster              Atom = 0x15706
	Pre                 Atom = 0x25e03
	Preload             Atom = 0x25e07
	Progress            Atom = 0x52c08
	Prompt              Atom = 0x5fa06
	Public              Atom = 0x41e06
	Q                   Atom = 0x13101
	Radiogroup          Atom = 0x30a
	Readonly            Atom = 0x2fb08
	Rel                 Atom = 0x25f03
	Required            Atom = 0x1d008
	Reversed            Atom = 0x5a08
	Rows                Atom = 0x9204
	Rowspan             Atom = 0x9207
	Rp                  Atom = 0x1c602
	Rt                  Atom = 0x13f02
	Ruby                Atom = 0xaf04
	S                   Atom = 0x2c01
	Samp                Atom = 0x4e04
	Sandbox             Atom = 0xbb07
	Scope               Atom = 0x2bd05
	Scoped              Atom = 0x2bd06
	Script              Atom = 0x3d406
	Seamless            Atom = 0x31c08
	Section             Atom = 0x4e207
	Select              Atom = 0x57a06
	Selected            Atom = 0x57a08
	Shape               Atom = 0x4f905
	Size                Atom = 0x55504
	Sizes               Atom = 0x55505
	Small               Atom = 0x18f05
	Sortable            Atom = 0x58d08
	Sorted              Atom = 0x19906
	Source              Atom = 0x1aa06
	Spacer              Atom = 0x2db06
	Span                Atom = 0x9504
	Spellcheck          Atom = 0x3230a
	Src                 Atom = 0x3c303
	Srcdoc              Atom = 0x3c306
	Srclang             Atom = 0x41107
	Start               Atom = 0x38605
	Step                Atom = 0x5f704
	Strike              Atom = 0x53306
	Strong              Atom = 0x55906
	Style               Atom = 0x61105
	Sub                 Atom = 0x5a903
	Summary             Atom = 0x61607
	Sup                 Atom = 0x61d03
	Svg                 Atom = 0x62003
	System              Atom = 0x62306
	Tabindex            Atom = 0x46308
	Table               Atom = 0x42d05
	Target              Atom = 0x24b06
	Tbody               Atom = 0x2e05
	Td                  Atom = 0x4702
	Template            Atom = 0x62608
	Textarea            Atom = 0x2f608
	Tfoot               Atom = 0x8c05
	Th                  Atom = 0x22e02
	Thead               Atom = 0x2d405
	Time                Atom = 0xdd04
	Title               Atom = 0xa105
	Tr                  Atom = 0x10502
	Track               Atom = 0x10505
	Translate           Atom = 0x14009
	Tt                  Atom = 0x5302
	Type                Atom = 0x21404
	Typemustmatch       Atom = 0x2140d
	U                   Atom = 0xb01
	Ul                  Atom = 0x8a02
	Usemap              Atom = 0x51106
	Value               Atom = 0x4005
	Var                 Atom = 0x11503
	Video               Atom = 0x28105
	Wbr                 Atom = 0x12103
	Width               Atom = 0x50705
	Wrap                Atom = 0x58704
	Xmp                 Atom = 0xc103
)

const hash0 = 0xc17da63e

const maxAtomLen = 19

var table = [1 << 9]Atom{
	0x1:   0x48a0b, // onmousemove
	0x2:   0x5e209, // onwaiting
	0x3:   0x1fa13, // onautocompleteerror
	0x4:   0x5fa06, // prompt
	0x7:   0x5eb07, // optimum
	0x8:   0x1604,  // mark
	0xa:   0x5ad07, // itemref
	0xb:   0x4fe0a, // onpageshow
	0xc:   0x57a06, // select
	0xd:   0x17b09, // draggable
	0xe:   0x3e03,  // nav
	0xf:   0x17507, // command
	0x11:  0xb01,   // u
	0x14:  0x2d507, // headers
	0x15:  0x44a08, // datalist
	0x17:  0x4e04,  // samp
	0x1a:  0x3fb09, // onkeydown
	0x1b:  0x55f08, // onscroll
	0x1c:  0x15003, // col
	0x20:  0x3c908, // itemprop
	0x21:  0x2780a, // http-equiv
	0x22:  0x61d03, // sup
	0x24:  0x1d008, // required
	0x2b:  0x25e07, // preload
	0x2c:  0x6040d, // onbeforeprint
	0x2d:  0x3600b, // ondragenter
	0x2e:  0x50902, // dt
	0x2f:  0x5a708, // onsubmit
	0x30:  0x27002, // hr
	0x31:  0x32f0d, // oncontextmenu
	0x33:  0x29c05, // image
	0x34:  0x50d07, // onpause
	0x35:  0x25906, // hgroup
	0x36:  0x7704,  // ping
	0x37:  0x57808, // onselect
	0x3a:  0x11303, // div
	0x3b:  0x1fa0e, // onautocomplete
	0x40:  0x2eb02, // mi
	0x41:  0x31c08, // seamless
	0x42:  0x2807,  // charset
	0x43:  0x8502,  // id
	0x44:  0x5200a, // onpopstate
	0x45:  0x3ef03, // del
	0x46:  0x2cb07, // marquee
	0x47:  0x3309,  // accesskey
	0x49:  0x8d06,  // footer
	0x4a:  0x44e04, // list
	0x4b:  0x2b005, // ismap
	0x51:  0x33804, // menu
	0x52:  0x2f04,  // body
	0x55:  0x9a08,  // frameset
	0x56:  0x54a07, // onreset
	0x57:  0x12705, // blink
	0x58:  0xa105,  // title
	0x59:  0x38807, // article
	0x5b:  0x22e02, // th
	0x5d:  0x13101, // q
	0x5e:  0x3cf04, // open
	0x5f:  0x2fa04, // area
	0x61:  0x44206, // onload
	0x62:  0xda04,  // font
	0x63:  0xd604,  // base
	0x64:  0x16207, // colspan
	0x65:  0x53707, // keytype
	0x66:  0x11e02, // dl
	0x68:  0x1b008, // fieldset
	0x6a:  0x2eb03, // min
	0x6b:  0x11503, // var
	0x6f:  0x2d506, // header
	0x70:  0x13f02, // rt
	0x71:  0x15008, // colgroup
	0x72:  0x23502, // mn
	0x74:  0x13a07, // onabort
	0x75:  0x3906,  // keygen
	0x76:  0x4c209, // onoffline
	0x77:  0x21f09, // challenge
	0x78:  0x2b203, // map
	0x7a:  0x2e902, // h4
	0x7b:  0x3b607, // onerror
	0x7c:  0x2e109, // maxlength
	0x7d:  0x2f505, // mtext
	0x7e:  0xbb07,  // sandbox
	0x7f:  0x58b06, // onsort
	0x80:  0x100a,  // malignmark
	0x81:  0x45d04, // meta
	0x82:  0x7b05,  // async
	0x83:  0x2a702, // h3
	0x84:  0x26702, // dd
	0x85:  0x27004, // href
	0x86:  0x6e0a,  // mediagroup
	0x87:  0x19406, // coords
	0x88:  0x41107, // srclang
	0x89:  0x34d0a, // ondblclick
	0x8a:  0x4005,  // value
	0x8c:  0xe908,  // oncancel
	0x8e:  0x3230a, // spellcheck
	0x8f:  0x9a05,  // frame
	0x91:  0x12403, // big
	0x94:  0x1f606, // action
	0x95:  0x6903,  // dir
	0x97:  0x2fb08, // readonly
	0x99:  0x42d05, // table
	0x9a:  0x61607, // summary
	0x9b:  0x12103, // wbr
	0x9c:  0x30a,   // radiogroup
	0x9d:  0x6c04,  // name
	0x9f:  0x62306, // system
	0xa1:  0x15d05, // color
	0xa2:  0x7f06,  // canvas
	0xa3:  0x25504, // html
	0xa5:  0x56f09, // onseeking
	0xac:  0x4f905, // shape
	0xad:  0x25f03, // rel
	0xae:  0x28510, // oncanplaythrough
	0xaf:  0x3760a, // ondragover
	0xb0:  0x62608, // template
	0xb1:  0x1d80d, // foreignObject
	0xb3:  0x9204,  // rows
	0xb6:  0x44e07, // listing
	0xb7:  0x49c06, // output
	0xb9:  0x3310b, // contextmenu
	0xbb:  0x11f03, // low
	0xbc:  0x1c602, // rp
	0xbd:  0x5bb09, // onsuspend
	0xbe:  0x13606, // button
	0xbf:  0x4db04, // desc
	0xc1:  0x4e207, // section
	0xc2:  0x52a0a, // onprogress
	0xc3:  0x59e09, // onstorage
	0xc4:  0x2d204, // math
	0xc5:  0x4503,  // alt
	0xc7:  0x8a02,  // ul
	0xc8:  0x5107,  // pattern
	0xc9:  0x4b60c, // onmousewheel
	0xca:  0x35709, // ondragend
	0xcb:  0xaf04,  // ruby
	0xcc:  0xc01,   // p
	0xcd:  0x31707, // onclose
	0xce:  0x24205, // meter
	0xcf:  0x11807, // bgsound
	0xd2:  0x25106, // height
	0xd4:  0x101,   // b
	0xd5:  0x2c308, // itemtype
	0xd8:  0x1bb07, // caption
	0xd9:  0x10c08, // disabled
	0xdb:  0x33808, // menuitem
	0xdc:  0x62003, // svg
	0xdd:  0x18f05, // small
	0xde:  0x44a04, // data
	0xe0:  0x4cb08, // ononline
	0xe1:  0x2a206, // mglyph
	0xe3:  0x6505,  // embed
	0xe4:  0x10502, // tr
	0xe5:  0x46b0b, // onloadstart
	0xe7:  0x3c306, // srcdoc
	0xeb:  0x5c408, // ontoggle
	0xed:  0xe703,  // bdo
	0xee:  0x4702,  // td
	0xef:  0x8305,  // aside
	0xf0:  0x29402, // h2
	0xf1:  0x52c08, // progress
	0xf2:  0x12c0a, // blockquote
	0xf4:  0xf005,  // label
	0xf5:  0x601,   // i
	0xf7:  0x9207,  // rowspan
	0xfb:  0x51709, // onplaying
	0xfd:  0x2a103, // img
	0xfe:  0xf608,  // optgroup
	0xff:  0x42307, // content
	0x101: 0x53e0c, // onratechange
	0x103: 0x3da0c, // onhashchange
	0x104: 0x4807,  // details
	0x106: 0x40008, // download
	0x109: 0x14009, // translate
	0x10b: 0x4230f, // contenteditable
	0x10d: 0x36b0b, // ondragleave
	0x10e: 0x2106,  // accept
	0x10f: 0x57a08, // selected
	0x112: 0x1f20a, // formaction
	0x113: 0x5b506, // center
	0x115: 0x45510, // onloadedmetadata
	0x116: 0x12804, // link
	0x117: 0xdd04,  // time
	0x118: 0x19f0b, // crossorigin
	0x119: 0x3bd07, // onfocus
	0x11a: 0x58704, // wrap
	0x11b: 0x42204, // icon
	0x11d: 0x28105, // video
	0x11e: 0x4de05, // class
	0x121: 0x5d40e, // onvolumechange
	0x122: 0xaa06,  // onblur
	0x123: 0x2b909, // itemscope
	0x124: 0x61105, // style
	0x127: 0x41e06, // public
	0x129: 0x2320e, // formnovalidate
	0x12a: 0x58206, // onshow
	0x12c: 0x51706, // onplay
	0x12d: 0x3c804, // cite
	0x12e: 0x2bc02, // ms
	0x12f: 0xdb0c,  // ontimeupdate
	0x130: 0x10904, // kind
	0x131: 0x2470a, // formtarget
	0x135: 0x3af07, // onended
	0x136: 0x26506, // hidden
	0x137: 0x2c01,  // s
	0x139: 0x2280a, // formmethod
	0x13a: 0x3e805, // input
	0x13c: 0x50b02, // h6
	0x13d: 0xc902,  // ol
	0x13e: 0x3420b, // oncuechange
	0x13f: 0x1e50d, // foreignobject
	0x143: 0x4e70e, // onbeforeunload
	0x144: 0x2bd05, // scope
	0x145: 0x39609, // onemptied
	0x146: 0x14b05, // defer
	0x147: 0xc103,  // xmp
	0x148: 0x39f10, // ondurationchange
	0x149: 0x1903,  // kbd
	0x14c: 0x47609, // onmessage
	0x14d: 0x60006, // option
	0x14e: 0x2eb09, // minlength
	0x14f: 0x32807, // checked
	0x150: 0xce08,  // autoplay
	0x152: 0x202,   // br
	0x153: 0x2360a, // novalidate
	0x156: 0x6307,  // noembed
	0x159: 0x31007, // onclick
	0x15a: 0x47f0b, // onmousedown
	0x15b: 0x3a708, // onchange
	0x15e: 0x3f209, // oninvalid
	0x15f: 0x2bd06, // scoped
	0x160: 0x18808, // controls
	0x161: 0x30b05, // muted
	0x162: 0x58d08, // sortable
	0x163: 0x51106, // usemap
	0x164: 0x1b80a, // figcaption
	0x165: 0x35706, // ondrag
	0x166: 0x26b04, // high
	0x168: 0x3c303, // src
	0x169: 0x15706, // poster
	0x16b: 0x1670e, // annotation-xml
	0x16c: 0x5f704, // step
	0x16d: 0x4,     // abbr
	0x16e: 0x1b06,  // dialog
	0x170: 0x1202,  // li
	0x172: 0x3ed02, // mo
	0x175: 0x1d803, // for
	0x176: 0x1a803, // ins
	0x178: 0x55504, // size
	0x179: 0x43210, // onlanguagechange
	0x17a: 0x8607,  // default
	0x17b: 0x1a03,  // bdi
	0x17c: 0x4d30a, // onpagehide
	0x17d: 0x6907,  // dirname
	0x17e: 0x21404, // type
	0x17f: 0x1f204, // form
	0x181: 0x28509, // oncanplay
	0x182: 0x6103,  // dfn
	0x183: 0x46308, // tabindex
	0x186: 0x6502,  // em
	0x187: 0x27404, // lang
	0x189: 0x39108, // dropzone
	0x18a: 0x4080a, // onkeypress
	0x18b: 0x23c08, // datetime
	0x18c: 0x16204, // cols
	0x18d: 0x1,     // a
	0x18e: 0x4420c, // onloadeddata
	0x190: 0xa605,  // audio
	0x192: 0x2e05,  // tbody
	0x193: 0x22c06, // method
	0x195: 0xf404,  // loop
	0x196: 0x29606, // iframe
	0x198: 0x2d504, // head
	0x19e: 0x5f108, // manifest
	0x19f: 0xb309,  // autofocus
	0x1a0: 0x14904, // code
	0x1a1: 0x55906, // strong
	0x1a2: 0x30308, // multiple
	0x1a3: 0xc05,   // param
	0x1a6: 0x21107, // enctype
	0x1a7: 0x5b304, // face
	0x1a8: 0xfd09,  // plaintext
	0x1a9: 0x26e02, // h1
	0x1aa: 0x59509, // onstalled
	0x1ad: 0x3d406, // script
	0x1ae: 0x2db06, // spacer
	0x1af: 0x55108, // onresize
	0x1b0: 0x4a20b, // onmouseover
	0x1b1: 0x5cc08, // onunload
	0x1b2: 0x56708, // onseeked
	0x1b4: 0x2140d, // typemustmatch
	0x1b5: 0x1cc06, // figure
	0x1b6: 0x4950a, // onmouseout
	0x1b7: 0x25e03, // pre
	0x1b8: 0x50705, // width
	0x1b9: 0x19906, // sorted
	0x1bb: 0x5704,  // nobr
	0x1be: 0x5302,  // tt
	0x1bf: 0x1105,  // align
	0x1c0: 0x3e607, // oninput
	0x1c3: 0x41807, // onkeyup
	0x1c6: 0x1c00c, // onafterprint
	0x1c7: 0x210e,  // accept-charset
	0x1c8: 0x33c06, // itemid
	0x1c9: 0x3e809, // inputmode
	0x1cb: 0x53306, // strike
	0x1cc: 0x5a903, // sub
	0x1cd: 0x10505, // track
	0x1ce: 0x38605, // start
	0x1d0: 0xd608,  // basefont
	0x1d6: 0x1aa06, // source
	0x1d7: 0x18206, // legend
	0x1d8: 0x2d405, // thead
	0x1da: 0x8c05,  // tfoot
	0x1dd: 0x1ec06, // object
	0x1de: 0x6e05,  // media
	0x1df: 0x1670a, // annotation
	0x1e0: 0x20d0b, // formenctype
	0x1e2: 0x3d208, // noscript
	0x1e4: 0x55505, // sizes
	0x1e5: 0x1fc0c, // autocomplete
	0x1e6: 0x9504,  // span
	0x1e7: 0x9808,  // noframes
	0x1e8: 0x24b06, // target
	0x1e9: 0x38f06, // ondrop
	0x1ea: 0x2b306, // applet
	0x1ec: 0x5a08,  // reversed
	0x1f0: 0x2a907, // isindex
	0x1f3: 0x27008, // hreflang
	0x1f5: 0x2f302, // h5
	0x1f6: 0x4f307, // address
	0x1fa: 0x2e103, // max
	0x1fb: 0xc30b,  // placeholder
	0x1fc: 0x2f608, // textarea
	0x1fe: 0x4ad09, // onmouseup
	0x1ff: 0x3800b, // ondragstart
}

const atomText = "abbradiogrouparamalignmarkbdialogaccept-charsetbodyaccesskey" +
	"genavaluealtdetailsampatternobreversedfnoembedirnamediagroup" +
	"ingasyncanvasidefaultfooterowspanoframesetitleaudionblurubya" +
	"utofocusandboxmplaceholderautoplaybasefontimeupdatebdoncance" +
	"labelooptgrouplaintextrackindisabledivarbgsoundlowbrbigblink" +
	"blockquotebuttonabortranslatecodefercolgroupostercolorcolspa" +
	"nnotation-xmlcommandraggablegendcontrolsmallcoordsortedcross" +
	"originsourcefieldsetfigcaptionafterprintfigurequiredforeignO" +
	"bjectforeignobjectformactionautocompleteerrorformenctypemust" +
	"matchallengeformmethodformnovalidatetimeterformtargetheightm" +
	"lhgroupreloadhiddenhigh1hreflanghttp-equivideoncanplaythroug" +
	"h2iframeimageimglyph3isindexismappletitemscopeditemtypemarqu" +
	"eematheaderspacermaxlength4minlength5mtextareadonlymultiplem" +
	"utedonclickoncloseamlesspellcheckedoncontextmenuitemidoncuec" +
	"hangeondblclickondragendondragenterondragleaveondragoverondr" +
	"agstarticleondropzonemptiedondurationchangeonendedonerroronf" +
	"ocusrcdocitempropenoscriptonhashchangeoninputmodeloninvalido" +
	"nkeydownloadonkeypressrclangonkeyupublicontenteditableonlang" +
	"uagechangeonloadeddatalistingonloadedmetadatabindexonloadsta" +
	"rtonmessageonmousedownonmousemoveonmouseoutputonmouseoveronm" +
	"ouseuponmousewheelonofflineononlineonpagehidesclassectionbef" +
	"oreunloaddresshapeonpageshowidth6onpausemaponplayingonpopsta" +
	"teonprogresstrikeytypeonratechangeonresetonresizestrongonscr" +
	"ollonseekedonseekingonselectedonshowraponsortableonstalledon" +
	"storageonsubmitemrefacenteronsuspendontoggleonunloadonvolume" +
	"changeonwaitingoptimumanifestepromptoptionbeforeprintstylesu" +
	"mmarysupsvgsystemplate"
