package integration

var (
	redis         = "docker.io/library/redis:alpine"
	fedoraMinimal = "registry.fedoraproject.org/fedora-minimal:latest"
	ALPINE        = "docker.io/library/alpine:latest"
	infra         = "k8s.gcr.io/pause:3.1"
	BB            = "docker.io/library/busybox:latest"
)
